#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>

/************************************************************************/
/* Debug and output symbols */
/************************************************************************/
#define DEBUG		0
#define DEEP_DEBUG	0
#define OUTPUT_RP	0

/************************************************************************/
/* Defaults */
/************************************************************************/
#define	DEF_MIN_ROLES			5
#define	DEF_MAX_ROLES			20
#define DEF_MIN_P_PER_ROLE		5
#define DEF_MAX_P_PER_ROLE		25
#define DEF_PERM_SHARING		50
#define DEF_MIN_CONSTR_RATIO		50
#define DEF_MAX_CONSTR_RATIO		2
#define DEF_MIN_R_PER_CONSTR		10
#define DEF_MIN_K_PER_CONSTR		3
#define DEF_SESSIONS			10
#define	DEF_SESS_ROLES			50		
#define DEF_OBJ				0
#define DEF_RH				0

#define DNAME_LEN			20

/************************************************************************/
/* Data structures */
/************************************************************************/
typedef struct pListType {
    int p; /* perm */
    struct pListType *next;
} pListType;

typedef struct rListType {
    int r; /* role */
    struct rListType *next;
} rListType;

typedef struct rpType {
    int r; /* Role id */
    int offset; /* To be used after pruning unwanted roles in sessions */
    pListType *pList; /* Perms assigned to the role -- adjacency list */
} rpType;

typedef struct rpListType {
    rpType rp;
    struct rpListType *next;
} rpListType;

typedef struct prType {
    int p; /* perm id */
    int offset; /* See above */
    rListType *rList; /* Roles to which this perm is assigned */
} prType;

typedef struct prListType {
    prType pr;
    struct prListType *next;
} prListType;

typedef struct constrType {
    rListType *r; /* Roles in constraint */
    int k;
} constrType;

typedef struct constrListType {
    constrType c;
    struct constrListType *next;
} constrListType;


/************************************************************************/
/* Global variables */
/************************************************************************/
int randFd = 0; /* fd for /dev/urandom */
const char *dnamePrefix = "pol-";
char dname[DNAME_LEN];

/************************************************************************/
/* Functions */
/************************************************************************/
void initRand() {
    randFd = open("/dev/urandom", O_RDONLY);
    if(randFd < 0) {
	perror("initRand()"); exit(0);
    }
}

unsigned int getRand() {
    unsigned int ret = 0;
    int retFromRead;

    retFromRead = read(randFd, &ret, sizeof(unsigned int));
    if(retFromRead < 0) {
	perror("getRand()"); exit(0);
    }
    else if(retFromRead < sizeof(unsigned int)) {
	fprintf(stderr, "Warning: getRand() -- only %d random bytes, not 4.\n", retFromRead);
    }

#if DEEP_DEBUG
    printf("genRand() returning: %u\n", ret);
#endif

    return ret;
}

void usage(char *progName) {
    printf("\n");
    printf("\tAll arguments are optional. Defaults are shown after the \'/\'\n");
    printf("\tUsage: %s <arguments>. Possible arguments, in order are:\n", progName);
    printf("\t\t-a: min # roles, default: %d\n", DEF_MIN_ROLES);
    printf("\t\t-b: max # roles, default: %d\n", DEF_MAX_ROLES);
    printf("\t\t-c: min # perms/role, default: %d\n", DEF_MIN_P_PER_ROLE);
    printf("\t\t-d: max # perms/role, default: %d\n", DEF_MAX_P_PER_ROLE);
    printf("\t\t-e: perm sharing among roles, between 0 & 100.\n");
    printf("\t\t    0 means no sharing, 100 means all perms shared.\n");
    printf("\t\t    Default: %d.\n", DEF_PERM_SHARING);
    printf("\t\t-f: min # constraints, default: # roles/%d\n", DEF_MIN_CONSTR_RATIO);
    printf("\t\t-g: max # constraints, default: # roles/%d\n", DEF_MAX_CONSTR_RATIO);
    printf("\t\t-h: min roles/constraint, default: %d\n", DEF_MIN_R_PER_CONSTR);
#if 1
    printf("\t\t-i: max roles/constraint, default: # roles/2}\n");
#endif
    printf("\t\t-j: min k in constraint, default: %d\n", DEF_MIN_K_PER_CONSTR);
#if 1
    printf("\t\t-k: max k in constraint, default: # roles in constr\n");
#endif
    printf("\t\t-l: num sessions to generate, default: %d\n", DEF_SESSIONS);
    printf("\t\t-m: %% roles in each session, default: %d\n", DEF_SESS_ROLES);
    printf("\t\t-n: num plb, default: random subset from -m option.\n");
    printf("\t\t-o: num pub, default: as many as for roles from -m option.\n");
    printf("\t\t-p: objective 0 (min) or 1 (max), default: %d\n", DEF_OBJ);
    printf("\t\t-q: max roles/perm, default: infinity.\n");
    printf("\t\t-r: RH depth, default: 0.\n");
    printf("\t\tNOTES:\n");
    printf("\t\t   1. # perms / role is constant\n");
    printf("\t\t   2. set of roles in constraints may be repeated\n");
    printf("\t\t   3. if -j/-k exceeds # roles in constr, it is made =\n");
    printf("\t\t   4. max roles per constr = # roles/2\n");
    printf("\t\t   5. max k per constr = # roles in constr\n");
    printf("\n");
}

void swap(int *a, int *b) {
    int tmp = *a;
    *a = *b;
    *b = tmp;
}

static int cmpInts(const void *p1, const void *p2) {
    const int *i1 = (const int *)p1;
    const int *i2 = (const int *)p2;

    if((*i1) < (*i2)) return 1;
    else if((*i1) == (*i2)) return 0;
    else return -1;
}

void shuffle(int *a, int alen) {
    int i;

    for(i = alen-1; i >= 0; i--) {
	int j = getRand()%(i+1);
	swap(&(a[i]), &(a[j]));
    }
}

void dumpRbacOutput(int numRoles,
	            int numPerms,
		    rpListType *rpl,
	            prListType *prl,
		    constrListType *cl) {

    const int fnameLen = DNAME_LEN + 20;
    char fname[fnameLen];

    strcpy(fname, dname);
    strcat(fname, "/rbac");

    FILE *fp = fopen(fname, "w");
    if(fp == NULL) {
	perror("fopen"); exit(0);
    }

    fprintf(fp, "#R\n%d\n\n", numRoles);
    fprintf(fp, "#P\n%d\n\n", numPerms);

    fprintf(fp, "#RP\n");
    rpListType *tmpRpList;
    for(tmpRpList = rpl; tmpRpList != NULL; tmpRpList = tmpRpList->next) {
	fprintf(fp, "%d ", (tmpRpList->rp).r);
	pListType *pl = (tmpRpList->rp).pList;
	for(; pl != NULL; pl = pl->next) {
	    fprintf(fp, "%d ", pl->p);
	}
	fprintf(fp, "\n");
    }

    fprintf(fp, "\n#PR\n");
    prListType *tmpPrList;
    for(tmpPrList = prl; tmpPrList != NULL; tmpPrList = tmpPrList->next) {
	fprintf(fp, "%d ", (tmpPrList->pr).p);
	rListType *rl = (tmpPrList->pr).rList;
	for(; rl != NULL; rl = rl->next) {
	    fprintf(fp, "%d ", rl->r);
	}
	fprintf(fp, "\n");
    }
    fprintf(fp, "\n#D\n");

    constrListType *tmpCl;
    for(tmpCl = cl; tmpCl != NULL; tmpCl = tmpCl->next) {
	rListType *crl = (tmpCl->c).r;
	for(;crl != NULL; crl = crl->next) {
	    fprintf(fp, "%d ", crl->r);
	}
	fprintf(fp, "%d\n", (tmpCl->c).k);
    }

    fclose(fp);
}

void dumpSessionOutput(int session,
		       int sessRoles,
		       int sessPerms,
		       prListType *prl,
		       rpListType *rpl,
		       constrListType *cl,
	               pListType *plb,
		       pListType *pub_plb,
		       int nOutputPlb,
		       int nOutputPub,
		       int maxRolesPerPerm,
		       int obj,
		       int rhDepth)
{
    const int fnameLen = DNAME_LEN + 20;
    char fname[fnameLen];

    sprintf(fname, "%s/session-%d.txt", dname, session);

    FILE *fp = fopen(fname, "w");
    if(fp == NULL) {
	perror("fopen"); exit(0);
    }

    fprintf(fp, "#R\n%d\n", sessRoles);

    fprintf(fp, "\n#P\n%d\n", sessPerms);

    fprintf(fp, "\n#PR\n");
    prListType *tprl = prl;
    for(; tprl != NULL; tprl = tprl->next) {
	if((tprl->pr).offset == -1) continue;

	fprintf(fp, "%d ", (tprl->pr).p - (tprl->pr).offset);
	rListType *rl = (tprl->pr).rList;
	rpListType *trpl = rpl;
	int numRolesThisPerm = 0;

	for(; rl != NULL; rl = rl->next) {
	    for(; (trpl->rp).r < rl->r; trpl = trpl->next) ;
	    /* (trpl->rp).r == rl->r */
	    if((trpl->rp).offset != -1) {
		// fprintf(fp, "%d ", (trpl->rp).r - (trpl->rp).offset);
		numRolesThisPerm++;
	    }
	}

	int rolesThisPerm[numRolesThisPerm];
	rl = (tprl->pr).rList;
	trpl = rpl;
	int i;
	for(i = 0; rl != NULL; rl = rl->next) {
	    for(; (trpl->rp).r < rl->r; trpl = trpl->next) ;
	    /* (trpl->rp).r == rl->r */
	    if((trpl->rp).offset != -1) {
		// fprintf(fp, "%d ", (trpl->rp).r - (trpl->rp).offset);
		rolesThisPerm[i++] = (trpl->rp).r - (trpl->rp).offset;
	    }
	}

	if(numRolesThisPerm < maxRolesPerPerm) {
	    /* Just output them */
	    int j;
	    for(j = 0; j < numRolesThisPerm; j++) {
		fprintf(fp, "%d ", rolesThisPerm[j]);
	    }
	}
	else {
	    shuffle(rolesThisPerm, numRolesThisPerm);
	    qsort(rolesThisPerm, maxRolesPerPerm, sizeof(int), cmpInts);
	    int j;
	    for(j = maxRolesPerPerm - 1; j >= 0; j--) {
		fprintf(fp, "%d ", rolesThisPerm[j]);
	    }
	}

	fprintf(fp, "\n");
    }

#if OUTPUT_RP

    /* SERIOUS WARNING: need to fix this code
     * to account for maxRolesPerPerm */

    fprintf(fp, "\n#RP\n");
    rpListType *trpl = rpl;
    for(; trpl != NULL; trpl = trpl->next) {
	if((trpl->rp).offset == -1) continue;

	fprintf(fp, "%d ", (trpl->rp).r - (trpl->rp).offset);
	pListType *pl = (trpl->rp).pList;
        prListType *tprl = prl;

	for(; pl != NULL; pl = pl->next) {
	    for(; (tprl->pr).p < pl->p; tprl = tprl->next) ;
	    /* (tprl->pr).p == pl->p */
	    if((tprl->pr).offset != -1) {
		fprintf(fp, "%d ", (tprl->pr).p - (tprl->pr).offset);
	    }
	}
	fprintf(fp, "\n");
    }

#endif

    if(rhDepth > 0) {
	fprintf(fp, "\n#RH\n");

	/* Num roles at each depth = sessRoles/(rhDepth + 1) */
	int i, j, k;

	for(i = 0; i < rhDepth; i++) {
	    for(j = i*sessRoles/(rhDepth + 1);
		    j < (i+1)*(sessRoles/(rhDepth+1)); j++) {

		fprintf(fp, "%d ", j+1);
		for(k = (i+1)*(sessRoles/(rhDepth+1)); k < sessRoles; k++) {
		    /* Toss a coin for whether this role is a child */
		    if(getRand()%2) {
			fprintf(fp, "%d ", k+1);
		    }
		}
		fprintf(fp, "\n");
	    }
	}
    }

    fprintf(fp, "\n#D\n");
    constrListType *tcl = cl;
    for(; tcl != NULL; tcl = tcl->next) {
	rListType *rl = (tcl->c).r;
	rpListType *trpl = rpl;

	for(; rl != NULL; rl = rl->next) {
	    for(;  (trpl->rp).r < rl->r; trpl = trpl->next) ;
	    /* (trpl->rp).r == rl->r */
	    if((trpl->rp).offset != -1) {
		fprintf(fp, "%d ", (trpl->rp).r - (trpl->rp).offset);
	    }
	}
	fprintf(fp, "%d\n", (tcl->c).k);
    }

    fprintf(fp, "\n#PLB\n");
#if 0
    for(tprl = prl; tprl != NULL; tprl = tprl->next) {
	if((tprl->pr).offset != -1) {
	    fprintf(fp, "%d ", (tprl->pr).p - (tprl->pr).offset);
	}
    }
#else
    int nPlbInList = 0;
    pListType *tmpPlb = plb;
    tprl = prl;
    for(; tmpPlb != NULL; tmpPlb = tmpPlb->next) {
	for(; (tprl->pr).p < tmpPlb->p; tprl = tprl->next) ;
	/* (tprl->pr).p == plb->p */
	if((tprl->pr).offset == -1) {
	    /* BUG!! */
	    fprintf(stderr, "Fatal -- 313\n");
	    exit(0);
	}
	// fprintf(fp, "%d ", (tprl->pr).p - (tprl->pr).offset);
	nPlbInList++;
    }

    int plbAr[nPlbInList], i;
    tmpPlb = plb; tprl = prl;
    for(i = 0; tmpPlb != NULL; tmpPlb = tmpPlb->next) {
	for(; (tprl->pr).p < tmpPlb->p; tprl = tprl->next) ;
	/* (tprl->pr).p == plb->p */
	if((tprl->pr).offset == -1) {
	    /* BUG!! */
	    fprintf(stderr, "Fatal -- 313\n");
	    exit(0);
	}
	// fprintf(fp, "%d ", (tprl->pr).p - (tprl->pr).offset);
	plbAr[i++] = (tprl->pr).p - (tprl->pr).offset;
    }
    /* Need nOutputPlb of them randomly. So randomize first. */
    shuffle(plbAr, nPlbInList);

    int nPlbWeWant = 0;
    if(nOutputPlb < nPlbInList) {
	nPlbWeWant = nOutputPlb;
    }
    else {
	nPlbWeWant = nPlbInList;
    }
    qsort(plbAr, nPlbWeWant, sizeof(int), cmpInts);

    for(i = nPlbWeWant - 1; i >= 0; i--) {
	fprintf(fp, "%d ", plbAr[i]);
    }
#endif

    fprintf(fp, "\n");

    fprintf(fp, "\n#PUB-PLB\n");
#if 1
    nPlbInList = 0;
    tprl = prl; tmpPlb = pub_plb;
    for(; tmpPlb != NULL; tmpPlb = tmpPlb->next) {
 	for(; (tprl->pr).p < tmpPlb->p; tprl = tprl->next) ;
	/* (tprl->pr).p == pub_plb->p */
	if((tprl->pr).offset == -1) {
	    /* BUG!! */
	    fprintf(stderr, "Fatal -- 313\n");
	    exit(0);
	}
	//fprintf(fp, "%d ", (tprl->pr).p - (tprl->pr).offset);
	nPlbInList++;
    }

    int pubAr[nPlbInList];
    tprl = prl; tmpPlb = pub_plb;
    for(i = 0; tmpPlb != NULL; tmpPlb = tmpPlb->next) {
 	for(; (tprl->pr).p < tmpPlb->p; tprl = tprl->next) ;
	/* (tprl->pr).p == pub_plb->p */
	if((tprl->pr).offset == -1) {
	    /* BUG!! */
	    fprintf(stderr, "Fatal -- 313\n");
	    exit(0);
	}
	//fprintf(fp, "%d ", (tprl->pr).p - (tprl->pr).offset);
	pubAr[i++] = (tprl->pr).p - (tprl->pr).offset;
    }

    shuffle(pubAr, nPlbInList);
    nPlbWeWant = 0;
    if(nOutputPub - nOutputPlb < nPlbInList) {
	nPlbWeWant = nOutputPub - nOutputPlb;
    }
    else {
	nPlbWeWant = nPlbInList;
    }
    qsort(pubAr, nPlbWeWant, sizeof(int), cmpInts);

    for(i = nPlbWeWant - 1; i >= 0; i--) {
	fprintf(fp, "%d ", pubAr[i]);
    }
#endif
    fprintf(fp, "\n");

    fprintf(fp, "\n#OBJ\n");
    fprintf(fp, "%d\n", obj);

    fclose(fp);
}

pListType *reversePList(pListType *pl) {
    pListType *newPl = NULL;

    while(pl != NULL) {
	pListType *tmpPl = pl;
	pl = pl->next;

	tmpPl->next = newPl;
	newPl = tmpPl;
    }

    return(newPl);
}

int main(int argc, char *argv[]) {
    usage(argv[0]);
    initRand();

#if DEBUG
    int num = 3;
    int b[num], i;
    for(i = 1; i <= num; i++) {
	b[i-1] = i;
    }

#if DEEP_DEBUG
    for(i = 0; i < num; i++) {
	printf("%d ", b[i]);
    }
    printf("\n");
#endif

    shuffle(b, num);

    for(i = 0; i < num; i++) {
	printf("%d ", b[i]);
    }
    printf("\n");
#endif

    /* Variables for the arguments */
    int minRoles = DEF_MIN_ROLES,
	maxRoles = DEF_MAX_ROLES,
	roles;

    int rhDepth = DEF_RH;

    if(maxRoles < minRoles) maxRoles = minRoles;
    roles = (getRand()%(maxRoles - minRoles + 1)) + minRoles;

    int minPPerRole = DEF_MIN_P_PER_ROLE,
	maxPPerRole = DEF_MAX_P_PER_ROLE,
	pPerRole;

    if(maxPPerRole < minPPerRole) maxPPerRole = minPPerRole;
    pPerRole = (getRand()%(maxPPerRole - minPPerRole + 1)) + minPPerRole;

    int permSharing = DEF_PERM_SHARING;
    int minConstr, maxConstr, constr;

    minConstr = roles/DEF_MIN_CONSTR_RATIO;
    maxConstr = roles/DEF_MAX_CONSTR_RATIO;
    if(maxConstr < minConstr) maxConstr = minConstr;
    constr = (getRand()%(maxConstr - minConstr + 1)) + minConstr;

    int minRPerConstr, maxRPerConstr;

    minRPerConstr = DEF_MIN_R_PER_CONSTR;
    maxRPerConstr = roles/2;

    if(maxRPerConstr < minRPerConstr) maxRPerConstr = minRPerConstr;

    int minKPerConstr, maxKPerConstr;
    
    minKPerConstr = DEF_MIN_K_PER_CONSTR;
    maxKPerConstr = maxRPerConstr;
    if(maxKPerConstr < minKPerConstr) maxKPerConstr = minKPerConstr;

    int sessions = DEF_SESSIONS;
    int nSessRoles = (roles*DEF_SESS_ROLES)/100;
    int obj = DEF_OBJ;
    int nOutputPlb = 0, nOutputPub = 0;

    int maxRolesPerPerm = INT_MAX; /* Infinity */

    int opt;
    while((opt = getopt(argc, argv, "a:b:c:d:e:f:g:h:i:j:k:l:m:n:o:p:q:r:")) != -1) {
	switch(opt) {
	    case 'a':
		minRoles = atoi(optarg);
		if(maxRoles < minRoles) maxRoles = minRoles;
		roles = (getRand()%(maxRoles - minRoles + 1)) + minRoles;
		nSessRoles = (roles*DEF_SESS_ROLES)/100;
		minConstr = roles/DEF_MIN_CONSTR_RATIO;
		maxConstr = roles/DEF_MAX_CONSTR_RATIO;
		constr = (getRand()%(maxConstr - minConstr + 1)) + minConstr;
		break;
	    case 'b':
		maxRoles = atoi(optarg);
		if(maxRoles < minRoles) maxRoles = minRoles;
		roles = (getRand()%(maxRoles - minRoles + 1)) + minRoles;
		nSessRoles = (roles*DEF_SESS_ROLES)/100;
		minConstr = roles/DEF_MIN_CONSTR_RATIO;
		maxConstr = roles/DEF_MAX_CONSTR_RATIO;
		constr = (getRand()%(maxConstr - minConstr + 1)) + minConstr;
		break;
	    case 'c':
		minPPerRole = atoi(optarg);
		if(maxPPerRole < minPPerRole) maxPPerRole = minPPerRole;
		pPerRole = (getRand()%(maxPPerRole - minPPerRole + 1))
			    + minPPerRole;
		break;
	    case 'd':
		maxPPerRole = atoi(optarg);
		if(maxPPerRole < minPPerRole) maxPPerRole = minPPerRole;
		pPerRole = (getRand()%(maxPPerRole - minPPerRole + 1))
			    + minPPerRole;
		break;
	    case 'e':
		permSharing = atoi(optarg);
		if((permSharing < 0) || (permSharing > 100)) {
		    fprintf(stderr, "Invalid value for -e. See usage. Reverting to default.\n");
		    permSharing = DEF_PERM_SHARING;
		}
		break;
	    case 'f':
		minConstr = atoi(optarg);
		if(maxConstr < minConstr) maxConstr = minConstr;
		constr = (getRand()%(maxConstr - minConstr + 1)) + minConstr;
		break;
	    case 'g':
		maxConstr = atoi(optarg);
		if(maxConstr < minConstr) maxConstr = minConstr;
		constr = (getRand()%(maxConstr - minConstr + 1)) + minConstr;
		break;
	    case 'h':
		minRPerConstr = atoi(optarg);
		if(maxRPerConstr < minRPerConstr) maxRPerConstr = minRPerConstr;
		break;
#if 1
	    case 'i':
		maxRPerConstr = atoi(optarg);
		if(maxRPerConstr < minRPerConstr) maxRPerConstr = minRPerConstr;
		break;
#endif
	    case 'j':
		minKPerConstr = atoi(optarg);
		if(maxKPerConstr < minKPerConstr) maxKPerConstr = minKPerConstr;
		break;
#if 1
	    case 'k':
		maxKPerConstr = atoi(optarg);
		if(maxKPerConstr < minKPerConstr) maxKPerConstr = minKPerConstr;
		break;
#endif
	    case 'l':
		sessions = atoi(optarg);
		break;
	    case 'm':
		{
		    int j = atoi(optarg);
		    if(((j*roles)/100 < 1) || ((j*roles)/100 > roles)) {
			fprintf(stderr,
				"Invalid -m option. Keeping default.\n");
		    }
		    else nSessRoles = (roles*j)/100;
		}
		break;
	    case 'n':
		nOutputPlb = atoi(optarg);
		break;
	    case 'o':
		nOutputPub = atoi(optarg);
		if(nOutputPub < nOutputPlb) nOutputPub = nOutputPlb;
		break;
	    case 'p':
		{
		    int j = atoi(optarg);
		    if((j != 0) && (j != 1)) {
			fprintf(stderr,
				"Invalid objective. Keeping default.\n");
		    }
		    else obj = j;
		}
		break;
	    case 'q':
		maxRolesPerPerm = atoi(optarg);
		break;
	    case 'r':
		rhDepth = atoi(optarg);
		break;
	    default:
		fprintf(stderr, "Error in options. See usage. Exiting...\n");
		exit(0);
		break;
	}
    }

#if 0
    maxKPerConstr = roles/2;
    maxRPerConstr = roles/2;
#endif
    if(maxRPerConstr < minRPerConstr)
	maxKPerConstr = maxRPerConstr = minRPerConstr;

    printf("\n");
    printf("\tGenerating policy with the following:\n");
    printf("\t\t# roles: %d\n", roles);
    printf("\t\t# perms per role: %d\n", pPerRole);
    printf("\t\tperm sharing: %d\n", permSharing);
    printf("\t\t# constraints: %d\n", constr);
    printf("\t\tmin/max roles per constr: %d/%d\n",
	    minRPerConstr, maxRPerConstr);
    printf("\t\tmin/max k per constr: %d/%d\n",
	    minKPerConstr, maxKPerConstr);
    printf("\t\tnum sessions: %d\n", sessions);
    printf("\t\tnSessRoles: %d\n", nSessRoles);
    printf("\t\tobjective: %d\n", obj);
    printf("\t\trhDepth: %d\n", rhDepth);
    printf("\n");

    /* Prepare for output */

    strcpy(dname, dnamePrefix);

    int i;
    for(i = 0; i < DNAME_LEN - strlen(dnamePrefix) - 1; i++) {
	char c = ((getRand()%10) + '0');
	strncat(dname, &c, 1);
    }

    if(mkdir(dname, 0777) < 0) {
	perror("mkdir"); exit(0);
    }

    printf("Output will go into dir: %s\n", dname);

    rpListType *rpl = NULL; /* Role -> perms */
    prListType *prl = NULL; /* Perms -> roles */

    /* Assume that roles are identified by [1, roles] */
    /* Assume that perms are identified by [roles+1, roles+# perms] */

    /* Initialize roles in rpl based on roles */
    for(i = roles; i > 0; i--) {
	rpListType *newRpl = (rpListType *)malloc(sizeof(rpListType));

	(newRpl->rp).r = i; /* Role id */
	(newRpl->rp).offset = 0; /* none yet */
	(newRpl->rp).pList = NULL; /* none yet */

	newRpl->next = rpl;
	rpl = newRpl;
    }

    /* Generate perms */
    int numPerms = 0; /* Num perms generated so far */
    int npToGenerate = (pPerRole * (100 - permSharing))/100;
    rpListType *tmpRpList = rpl;
    for(;tmpRpList != NULL; tmpRpList = tmpRpList->next) {
	int numNewPerms = 0;

	if(numPerms < (pPerRole - npToGenerate)) {
	    numNewPerms = pPerRole - numPerms;
	}
	else {
	    /* Generate npToGenerate new perms */
	    numNewPerms = npToGenerate;
	}

	/* Start to create a pList & prList */
	pListType *pl = NULL;
	prListType *newPrl = NULL;
	for(i = roles+numPerms+numNewPerms; i > roles+numPerms; i--) {
	    pListType *tmpPl = (pListType *)malloc(sizeof(pListType));
	    tmpPl->p = i;
	    tmpPl->next = pl;
	    pl = tmpPl;

	    prListType *tmpNewPrl = (prListType *)malloc(sizeof(prListType));
	    (tmpNewPrl->pr).p = i;
	    (tmpNewPrl->pr).offset = 0;
	    (tmpNewPrl->pr).rList = (rListType *)malloc(sizeof(rListType));
	    ((tmpNewPrl->pr).rList)->r = (tmpRpList->rp).r;
	    ((tmpNewPrl->pr).rList)->next = NULL;
	    tmpNewPrl->next = newPrl;
	    newPrl = tmpNewPrl;
	}

	/* Also need pPerRole - numNewPerms existing perms */
	/* Prepare array for random choices */
	if((pPerRole - numNewPerms) > 0) {
	    int a[numPerms];
	    for(i = 0; i < numPerms; i++) {
		a[i] = roles+i+1;
	    }

	    shuffle(a, numPerms);
	    qsort(a, pPerRole - numNewPerms, sizeof(int), cmpInts);

	    for(i = 0; i < pPerRole - numNewPerms; i++) {
		pListType *tmpPl = (pListType *)malloc(sizeof(pListType));
		tmpPl->p = a[i];
		tmpPl->next = pl;
		pl = tmpPl;
		
		/* Need to also add this role to the apt prl entry */
		prListType *tmpPrl = prl;
		for(; tmpPrl != NULL; tmpPrl = tmpPrl->next) {
		    if((tmpPrl->pr).p == a[i]) {
			/* Append this role id to the end of list of roles */
			rListType *existRl = (tmpPrl->pr).rList;
			
			if(existRl == NULL) {
			    fprintf(stderr, "Fatal -- existRl NULL 425\n");
			    exit(0);
			}

			for(;existRl->next != NULL; existRl = existRl->next) ;

			existRl->next = (rListType *)malloc(sizeof(rListType));
			existRl->next->r = (tmpRpList->rp).r;
			existRl->next->next = NULL;

			break;
		    }
		}

		if(tmpPrl == NULL) {
		    fprintf(stderr, "Fatal -- tmpPrl == NULL in line 438.\n");
		    exit(0);
		}
	    }
	}

	(tmpRpList->rp).pList = pl;

	/* Append newPrl to prl */
	if(prl == NULL) prl = newPrl;
	else {
	    prListType *tmpPrl = prl;
	    for(; tmpPrl->next != NULL; tmpPrl = tmpPrl->next) ;
	    tmpPrl->next = newPrl;
	}

	numPerms += numNewPerms;
    }

    /* Generate constraints */
    constrListType *cl = NULL;
    for(i = 0; i < constr; i++) {
	/* WARNING: we will go into an infinite loop if constr > # of
	 * subsets of roles */
	constrListType *newCl = (constrListType *)malloc(sizeof(constrListType));
	int nRolesInC = (getRand()%(maxRPerConstr - minRPerConstr + 1)) + (minRPerConstr);
	int kInC = (getRand()%(maxKPerConstr - minKPerConstr + 1)) + (minKPerConstr);
	if((nRolesInC > roles) || (kInC > nRolesInC)) {
	    /* Retry */
	    i--; continue;
	}

	/* Choose set of roles at random */
	int a[roles];
	int j;
	for(j = 0; j < roles; j++) {
	    a[j] = j+1;
	}
	shuffle(a, roles);
	qsort(a, nRolesInC, sizeof(int), cmpInts);
	rListType *newRl = NULL;
	for(j = 0; j < nRolesInC; j++) {
	    rListType *tmpRl = (rListType *)malloc(sizeof(rListType));

	    tmpRl->r = a[j];
	    tmpRl->next = newRl;
	    newRl = tmpRl;
	}

	(newCl->c).r = newRl;
	(newCl->c).k = kInC;
	newCl->next = cl;
	cl = newCl;
    }

#if 0
    dumpRbacOutput(roles, numPerms, rpl, prl, cl);
#endif

#if 0
    printf("Num perms: %d\n", numPerms);
#endif

    /* Create sessions */
    for(i = 0; i < sessions; i++) {
	printf("Session %d: ", i+1);

	/* Reset all offsets */
	rpListType *trpl = rpl;
	for(;trpl != NULL; trpl = trpl->next) {
	    (trpl->rp).offset = 0;
	}

	prListType *tprl = prl;
	for(;tprl != NULL; tprl = tprl->next) {
	    (tprl->pr).offset = 0;
	}

	int a[roles], j;
	for(j = 0; j < roles; j++) {
	    a[j] = j+1;
	}
	shuffle(a, roles);
	qsort(a, nSessRoles, sizeof(int), cmpInts);

#if 0
	printf("\tChosen roles: ");
	for(j = nSessRoles-1; j >= 0; j--) {
	    printf("%d ", a[j]);
	}
	printf("\n");
#endif

	pListType *pub = NULL, *plb = NULL;
	int npub = 0, nplb = 0;

	tprl = prl;
	for(; tprl != NULL; tprl = tprl->next) {
	    rListType *rl = (tprl->pr).rList;
	    for(;rl != NULL; rl = rl->next) {
		int j;
		for(j = nSessRoles-1; j >= 0; j--) {
		    if(a[j] == (rl->r)) {
			pListType *newP =
			    (pListType *)malloc(sizeof(pListType));
			newP->p = (tprl->pr).p;
			newP->next = pub;
			pub = newP;
			npub++;

#if 0
			printf("\tAdded perm %d to pub based on role %d.\n",
				newP->p, a[j]);
#endif

			break;
		    }
		}

		if(j >= 0) break;
	    }
	}

#if 0
	printf("\tnpub: %d, ", npub);
#endif
	nplb = (getRand()%npub) + 1;
#if 0
	printf("\tnplb: %d\n", nplb);
#endif

	pub = reversePList(pub);

	pListType *pl = pub;
#if 0
	printf("\tpub: ");
	for(; pl != NULL; pl = pl->next) {
	    printf("%d ", pl->p);
	}
	printf("\n");
#endif

	int b[npub];
	for(j=0,pl = pub;pl != NULL; pl = pl->next,j++) {
	    b[j] = pl->p;
	}
	shuffle(b, npub);
	qsort(b, nplb, sizeof(int), cmpInts);

	for(j = 0; j < nplb; j++) {
	    pl = (pListType *)malloc(sizeof(pListType));
	    pl->p = b[j];
	    pl->next = plb;
	    plb = pl;
	}

#if 0
	printf("\tplb: ");
	for(pl = plb; pl != NULL; pl = pl->next) {
	    printf("%d ", pl->p);
	}
	printf("\n");
#endif

	/* Exclude unneeded roles & perms in rpl & prl */
	tprl = prl; pl = pub;
	for(; tprl != NULL; tprl = tprl->next) {
	    if((pl != NULL) && ((tprl->pr).p == (pl->p))) {
		pl = pl->next;
		continue;
	    }
	    else if((pl != NULL) && ((tprl->pr).p > (pl->p))) {
		/* BUG!! */
		/* tprl can never be ahead of pl */
		fprintf(stderr, "Fatal -- Line 719\n");
		exit(0);
	    }
	    /* else */
	    (tprl->pr).offset = -1;
	    rListType *rl = (tprl->pr).rList;

#if 0
	    printf("Perm %d removed.\n", a[j]);
#endif

	    for(; rl != NULL; rl = rl->next) {
		rpListType *trpl = rpl;

		for(; trpl != NULL; trpl = trpl->next) {
		    if((trpl->rp).r == rl->r) {

			(trpl->rp).offset = -1;

#if 0
			printf("Role %d removed.\n", rl->r);
#endif

			break;
		    }
		}

		if(trpl == NULL) {
		    fprintf(stderr, "Fatal at 743, rl->r: %d\n", rl->r);
		    exit(0);
		}
	    }
	}

	/* Fix non-negative offsets in rpl & prl */
	int sessRoles = roles, sessPerms = numPerms;
	int currOffset = 0;
	trpl = rpl;
	for(; trpl != NULL; trpl = trpl->next) {
	    if((trpl->rp).offset == -1) {
		currOffset += 1;
	    }
	    else {
		(trpl->rp).offset = currOffset;
	    }
	}
	sessRoles -= currOffset;

#if 0
	printf("\tsessRoles: %d, ", sessRoles);
#endif

	if(sessRoles <= 0) {
	    /* BUG!! */
	    fprintf(stderr, "Fatal -- Line 766\n");
	    exit(0);
	}

	tprl = prl;
	for(; tprl != NULL; tprl = tprl->next) {
	    if((tprl->pr).offset == -1) {
		currOffset += 1;
	    }
	    else {
		(tprl->pr).offset = currOffset;
	    }
	}
	sessPerms -= currOffset;
	sessPerms += (roles - sessRoles);

#if 0
	printf("sessPerms: %d\n", sessPerms);
#endif

	/* Compute pub-plb, store in pub_plb */
	pListType *pub_plb = NULL;
	pListType *tplb = plb, *tpub = pub;
	for(; tpub != NULL; tpub = tpub->next) {
	    if((tplb != NULL) && (tplb->p == tpub->p)) {
		tplb = tplb->next;
		continue;
	    }
	    else if((tplb != NULL) && (tplb->p < tpub->p)) {
		/* BUG!! */
		fprintf(stderr, "Fatal -- 802\n");
		exit(0);
	    }
	    /* else */
	    pListType *newPl = (pListType *)malloc(sizeof(pListType));
	    newPl->p = tpub->p;
	    newPl->next = pub_plb;
	    pub_plb = newPl;
	}

	pub_plb = reversePList(pub_plb);

#if 0
	printf("\tpub_plb: ");
	for(pl = pub_plb; pl != NULL; pl = pl->next) {
	    printf("%d ", pl->p);
	}
	printf("\n");
#endif
	
	/* Remove invalid constraints */
	constrListType *tcl = cl, *prevTcl = NULL;
	int nCRemoved = 0;
	for(; tcl != NULL; tcl = tcl->next) {
	    rListType *rl = (tcl->c).r;
	    rpListType *trpl = rpl;
	    int rlLen = 0, invalidRs = 0;

	    for(; rl != NULL; rl = rl->next) {
		rlLen++;
		for(; (trpl != NULL) && ((trpl->rp).r < rl->r);
			trpl = trpl->next) ;

		if((trpl == NULL) || ((trpl->rp).r != rl->r)) {
		    /* BUG!! */
		    fprintf(stderr, "Fatal -- 832.\n");
		    exit(0);
		}

		if((trpl->rp).offset == -1) {
		    invalidRs++;
		}
	    }

	    if((rlLen - invalidRs) < (tcl->c).k) {
		/* Remove this c entry */

#if 0
		printf("\tC entry: %d < %d, removing.\n",
			    rlLen-invalidRs, (tcl->c).k); 
#endif
		nCRemoved++;

		/* WARNING -- memory leak in the following lines! */
		if(prevTcl == NULL) {
		    cl = tcl->next;
		}
		else {
		    prevTcl->next = tcl->next;
		}
	    }
	    else {
		prevTcl = tcl;
	    }
	}

	if(cl == NULL) {
	    printf("Warning: no constraints."); fflush(stdout);
	}
#if 0
	else {
	    printf("\t # constraints removed: %d\n", nCRemoved);
	}
#endif

	dumpSessionOutput(i+1, sessRoles, sessPerms, prl, rpl, cl,
		          plb, pub_plb, nOutputPlb, nOutputPub,
			  maxRolesPerPerm, obj, rhDepth);

	printf("...done.\n");
    }

    return 0;
}
